""" An interface to the server for the JobsManager
"""

import time
from datetime import datetime

from cloudcontrol.server.db import SObject

from cloudcontrol.common.jobs import JobsManagerInterface
from cloudcontrol.common.tql.db.tag import StaticTag, CallbackTag


class ServerJobsManagerInterface(JobsManagerInterface):

    TAG_ATTRIBUTES = ('title', 'status', 'state', 'owner', 'created', 'ended',
                      'attachments')

    def __init__(self, server):
        self._server = server
        self._jobs = {}  # Store each TQL object relative to a job

    def on_job_created(self, job):
        tql_object = SObject(job.id)
        tql_object.register(StaticTag('r', 'job'))
        self._server.db.register(tql_object)
        self._jobs[job.id] = tql_object

    def on_job_updated(self, job):
        tql_object = self._jobs[job.id]


        for name in self.TAG_ATTRIBUTES:
            value = getattr(job, name, None)
            # Convert values:
            if isinstance(value, bool):
                value = 'yes' if value else 'no'
            elif isinstance(value, datetime):
                value = str(int(time.mktime(value.timetuple())))
            elif isinstance(value, set):
                value = ' '.join(value)
            elif value is None:
                continue
            else:
                value = str(value)
            if name not in tql_object:
                tag = StaticTag(name, '')
                tql_object.register(tag)
            else:
                tag = tql_object[name]
            tag.value = value

        if 'duration' not in tql_object:
            def compute_duration():
                end = datetime.now() if job.ended is None else job.ended
                dt = end - job.created
                return dt.seconds + dt.days * 86400
            tql_object.register(CallbackTag('duration', compute_duration, ttl=0))

    def on_job_purged(self, job_id):
        tql_object = self._jobs[job_id]
        self._server.db.unregister(tql_object)
        del self._jobs[job_id]